"""
General functions for scraping data from Google search engine results pages.
"""

import re
import requests
import urllib.parse
import pandas as pd
import numpy as np
from requests_html import HTMLSession


def _get_source(url: str):
    """Return the source code for the provided URL.

    Args:
        url (string): URL of the page to scrape.

    Returns:
        response (object): HTTP response object from requests_html.
    """

    try:
        session = HTMLSession()
        response = session.get(url)

        if response.status_code == 200:
            return response
        elif response.status_code == 429:
            print('Error: Too many requests. Google has temporarily blocked you. Try again later.')
            exit()
        else:
            print('Error:' + response)
            exit()
    except requests.exceptions.RequestException as e:
        print(e)


def _get_site_results(url: str):
    """Return the source of a site:url search.

    Args:
        url: URL of page to append to site: query

    Returns:
        response (str): HTML of page.
    """

    try:
        query = urllib.parse.quote_plus(url)
        response = _get_source("https://www.google.co.uk/search?q=site%3A" + query)

        return response
    except requests.exceptions.RequestException as e:
        print(e)


def _parse_site_results(response: str):
    """Parse the HTML of a site:url query and return the number of pages "indexed".

    Args:
        response: HTML of site:url query.

    Returns:
        indexed: Number of pages "indexed".
    """

    try:
        if response.html.find("#result-stats", first=True):

            string = response.html.find("#result-stats", first=True).text
            if string:
                # Remove values in paretheses, i.e. (0.31 seconds)
                string = re.sub(r'\([^)]*\)', '', string)

                # Remove non-numeric characters
                string = re.sub('[^0-9]', '', string)

                return string
            else:
                return 0
    except requests.exceptions.RequestException as e:
        print(e)


def _count_indexed_pages(url: str):
    """Gets the site:url data, parses the response, and returns the number of "indexed" pages.

    Args:
        url: URL to use in site:url search.

    Returns:
        results (int): Number of pages "indexed".
    """

    response = _get_site_results(url)
    return _parse_site_results(response)


def get_indexed_pages(urls: list):
    """Loop through a series of URLs and run site:url searches, then return number of "indexed" pages.

    Args:
        urls (list): List of URLs.

    Returns:
        df (dataframe): Pandas dataframe containing URL and number of "indexed" pages.
    """

    data = []
    for site in urls:
        site_data = {'url': site, 'indexed_pages': _count_indexed_pages(site)}
        data.append(site_data)
    df = pd.DataFrame.from_records(data)
    df = df.sort_values(by='indexed_pages')
    return df


def _get_results(query: str):
    """Return the source of a search.

    Args:
        query: Search query term.

    Returns:
        response (str): HTML of page.
    """

    query = urllib.parse.quote_plus(query)
    response = _get_source("https://www.google.co.uk/search?q=" + query)

    return response


def _get_next_page(response, domain="google.co.uk"):
    """Get the URL for the next page of results."""

    css_identifier_next = "#pnnext"
    next_page_url = response.html.find(css_identifier_next, first=True).attrs['href']
    next_page = "https://www." + domain + next_page_url

    return next_page


def _parse_search_results(response):
    """Parses the Google Search engine results and returns a list of results.

    Note: This function is obviously dependent upon the source code in the Google results.
    Google obfuscates the source of the page to make it more difficult to extra information.
    Extraction classes change from time to time, so there is always a likelihood that this
    function will need to be adjusted with the new class or identifier details.
    In the event of the function failing, please raise a GitHub issue.

    Args:
        response: Response object containing the page source code.

    Returns:
        list: List of Google search results.
    """

    css_identifier_result = ".tF2Cxc"  # The class of the div containing each result, i.e. <div class="tF2Cxc">
    css_identifier_title = "h3"  # The element containing the title, i.e. <h3 class="...
    css_identifier_link = ".yuRUbf a"  # The class of the div containing the anchor, i.e. <div class="yuRUbf"><a ...
    css_identifier_text = ".VwiC3b"  # The class of the parent element containing the snippet <span>
    css_identifier_bold = ".VwiC3b span em"  # The class of the element containing the snippet <span><em>

    try:
        results = response.html.find(css_identifier_result)

        output = []

        for result in results:

            if result.find(css_identifier_text, first=True):
                text = result.find(css_identifier_text, first=True).text
            else:
                text = ''

            if result.find(css_identifier_title, first=True):
                title = result.find(css_identifier_title, first=True).text
            else:
                title = ''

            if result.find(css_identifier_link, first=True):
                link = result.find(css_identifier_link, first=True).attrs['href']
            else:
                link = ''

            # Extract bold text
            if result.find(css_identifier_bold, first=True):
                bold = result.find(css_identifier_bold, first=True).text.lower()
            else:
                bold = ''

            item = {
                'title': title,
                'link': link,
                'text': text,
                'bold': bold,
            }

            output.append(item)

        return output
    except requests.exceptions.RequestException as e:
        print(e)


def get_serps(query: str,
              output="dataframe",
              pages=1,
              domain="google.co.uk"):
    """Return the Google search results for a given query.

    Args:
        query (string): Query term to search Google for.
        output (string, optional): Optional output format (dataframe or dictionary).
        pages (int, optional): Optional number of pages to return.
        domain (string, optional): Optional Google domain (default is google.co.uk).

    Returns:
        results (dict): Results of query.
    """

    response = _get_results(query)
    results = _parse_search_results(response)
    next_page = _get_next_page(response)

    page = 1
    while page <= pages:
        if page > 1:
            response = _get_source(next_page)
            results = results + _parse_search_results(response)
            next_page = _get_next_page(response)
        page += 1

    if results:
        if output == "dataframe":
            df = pd.DataFrame.from_records(results)
            df.index = np.arange(1, len(df) + 1)
            df.index.names = ['position']
            return df.reset_index()
        else:
            return results



"""
Fetch data from the Google Search Console API.
"""

import sys
import pandas as pd
from google.oauth2 import service_account
from googleapiclient.discovery import build


def _connect(key: str):
    """Create a connection to the Google Search Console API and return service object.

    Args:
        key (string): Google Search Console JSON client secrets path.

    Returns:
        service (object): Google Search Console service object.
    """

    try:
        scope = ['https://www.googleapis.com/auth/webmasters']
        credentials = service_account.Credentials.from_service_account_file(key, scopes=scope)
        service = build('webmasters', 'v3', credentials=credentials)

        return service

    except Exception as e:
        print("Error: ", e)
        sys.exit(1)


def _get_response(service, site_url, payload):
    """Returns the rowLimit value from a Google Search Console API payload.

    Args:
        service (object): Google Search Console service object.
        site_url (string): Site URL for the Google Search Console property.
                           For domain properties use "sc-domain:example.com".
                           For other properties use "https://www.example.com".
        payload (dict): Google Search Console API payload.

    Returns:
        response (dict): API response dictionary
    """

    try:
        return service.searchanalytics().query(siteUrl=site_url, body=payload).execute()
    except Exception as e:
        return e


def _get_results(service, site_url, payload, results):
    """Returns a dictionary containing the Google Search Console API query results.

    Args:
        service (object): Google Search Console service object.
        site_url (string): Site URL for the Google Search Console property.
                           For domain properties use "sc-domain:example.com".
                           For other properties use "https://www.example.com".
        payload (dict): Google Search Console API payload.
        results (list): Python list to which to append the results.

    Returns:
        results (dict): Python dictionary of results to use to create a dataframe.

    """

    response = _get_response(service, site_url, payload)

    try:
        for row in response['rows']:
            data = {}

            for i in range(len(payload['dimensions'])):
                data[payload['dimensions'][i]] = row['keys'][i]

            data['clicks'] = row['clicks']
            data['impressions'] = row['impressions']
            data['ctr'] = round(row['ctr'] * 100, 2)
            data['position'] = round(row['position'], 2)
            results.append(data)

        return results
    except Exception as e:
        return None


def query_google_search_console(key: str, site_url: str, payload: dict, fetch_all=False):
    """Run a query on the Google Search Console API and return a dataframe of results.

    Args:
        key (object): JSON client secrets key file path.
        site_url (string): URL of Google Search Console property
        payload (dict): API query payload dictionary
        fetch_all (boolean, default=False): Set to True to return all results and ignore rowLimit and startRow if provided

    Return:
        df (dataframe): Pandas dataframe containing requested data.
    """

    service = _connect(key)
    results = []

    if fetch_all == False:
        results = _get_results(service, site_url, payload, results)
    else:
        maxrows = 10000
        startrow = 0
        complete = False

        while not complete:
            payload['rowLimit'] = maxrows
            payload['startRow'] = startrow
            result = _get_results(service, site_url, payload, results)

            if result is None:
                complete = True

            startrow += maxrows
    return pd.DataFrame.from_dict(results)


def query_google_search_console_compare(key, site_url, payload_before, payload_after, fetch_all=False):
    """Run two queries on the Google Search Console API and return a dataframe of results comparing changes.

    Args:
        key (object): JSON client secrets key file path.
        site_url (string): URL of Google Search Console property
        payload_before (dict): API query payload dictionary for earliest period
        payload_after (dict): API query payload dictionary for latest period
        fetch_all (boolean, default=False): Set to True to return all results and ignore rowLimit and startRow if provided

    Example:

    The below code will compare queries by page and device across two periods and return a dataframe of results.

    payload_before = {
    'startDate': "2021-08-11",
    'endDate': "2021-08-31",
    'dimensions': ["page","query", "device"],
    }

    payload_after = {
        'startDate': "2021-07-21",
        'endDate': "2021-08-10",
        'dimensions': ["page","query", "device"],
    }

    df = query_google_search_console_compare(key, site_url, payload_before, payload_after, fetch_all=False)

    Return:
        df (dataframe): Pandas dataframe containing requested data.
    """

    # Validate the payload
    if ('date' in payload_before['dimensions']) or ('date' in payload_after['dimensions']):
        print('The date dimension cannot be used in a payload. Please use only page, query, and device.')
    elif payload_before['dimensions'] != payload_after['dimensions']:
        print('The payload dimensions provided do not match. Please use the same dimensions in each payload.')
    else:
        # Fetch the data and prefix the column names with _before and _after
        df_before = query_google_search_console(key, site_url, payload_before, fetch_all=fetch_all)
        df_after = query_google_search_console(key, site_url, payload_after, fetch_all=fetch_all)
        df_before.columns = [str(col) + '_before' for col in df_before.columns]
        df_after.columns = [str(col) + '_after' for col in df_after.columns]

        # Extract the dimensions from the payload, remove date and append _before and _after and join data
        dimensions_before = [dimension + '_before' for dimension in payload_before['dimensions']]
        dimensions_after = [dimension + '_after' for dimension in payload_after['dimensions']]
        df = df_before.merge(df_after, how='left', left_on=dimensions_before, right_on=dimensions_after)
        df = df.fillna(0)

        # Calculate changes between the periods
        df['clicks_change'] = df['clicks_after'] - df['clicks_before']
        df['impressions_change'] = df['impressions_after'] - df['impressions_before']
        df['ctr_change'] = df['ctr_after'] - df['ctr_before']
        df['position_change'] = df['position_after'] - df['position_before']

        # Drop the _after suffixed columns from the dataframe
        object_columns = list(df.select_dtypes(['object']).columns)
        after_columns = [column for column in object_columns if "_after" in column]
        df = df.drop(columns=after_columns)

        # Create the dataframe
        dimension_columns = list(df.select_dtypes(['object']).columns)
        metrics = ['impressions_before',
                   'impressions_after',
                   'impressions_change',
                    'clicks_before',
                   'clicks_after',
                   'clicks_change',
                   'ctr_before',
                   'ctr_after',
                   'ctr_change',
                   'position_before',
                   'position_after',
                   'position_change']
        df = df[dimension_columns + metrics]

        # Drop the _before from dimension columns
        df.columns = df.columns.str.replace('page_before', 'page')
        df.columns = df.columns.str.replace('query_before', 'query')
        df.columns = df.columns.str.replace('device_before', 'device')

        return df



"""
Functions for running simple before and after tests using Causal Impact.
"""

from datetime import timedelta
import pandas as pd
from causalimpact import CausalImpact
from ecommercetools import seo
import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter("ignore")


def _subtract_days_from_date(date, days):
    """Subtract days from a date and return the date.

    Args:
        date (string): Date string in YYYY-MM-DD format.
        days (int): Number of days to subtract from date

    Returns:
        date (date): Date in YYYY-MM-DD with X days subtracted.
    """

    subtracted_date = pd.to_datetime(date) - timedelta(days=days)
    subtracted_date = subtracted_date.strftime("%Y-%m-%d")

    return subtracted_date


def _add_days_to_date(date, days):
    """Add days to a date and return the date.

    Args:
        date (string): Date string in YYYY-MM-DD format.
        days (int): Number of days to add to date

    Returns:
        date (date): Date in YYYY-MM-DD with X days added.
    """

    added_date = pd.to_datetime(date) + timedelta(days=days)
    added_date = added_date.strftime("%Y-%m-%d")

    return added_date


def _get_pre_and_post_periods(post_period_start_date, days):
    """Return the pre- and post-period dates for use in CausalImpact.

    If you provide the start date for the test period, i.e. 2021-07-18, and
    the test duration in days, i.e. 14, this function will return the start
    and end date for the test period, and the start and end date for the
    pre-intervention period that ran immediately before. Data are returned
    in the right format for use by the CausalImpact model.

    Args:
        post_period_start_date (string): The date at which the test was started in YYYY-MM-DD format.
        days (int): The number of days to use for the test period.

    Returns:
        pre_period (list): The start and end date for the period before the test.
        post_period (list): The start and end date for the test period.

    Example:
        pre_period, post_period = get_pre_and_post_periods('2021-07-18', 14)

        pre_period
            ['2021-07-04', '2021-07-17']

        post_period
            ['2021-07-18', '2021-07-31']
    """

    pre_period_start_date = _subtract_days_from_date(post_period_start_date, days)
    pre_period_end_date = _subtract_days_from_date(post_period_start_date, 1)
    post_period_end_date = _add_days_to_date(post_period_start_date, days-1)

    pre_period = [pre_period_start_date, pre_period_end_date]
    post_period = [post_period_start_date, post_period_end_date]

    return pre_period, post_period


def _get_seo_test_data(key, site_url, post_period_start_date, days, filters=None):
    """Return Google Search Console data for use within a Causal Impact SEO test.

    Args:
        key (string): Filepath of Google Search Console API client secrets JSON keyfile
        site_url (string): Google Search Console property URL to query
        post_period_start_date (string): Start date for test in YYYY-MM-DD format
        days (int): Number of days to include in test period
        filters (optional, list): Optional list of GSC formatted query filters.

    Returns:
        df (dataframe): Date indexed dataframe containing clicks, impressions, ctr, and position

    Usage:
        # Site level test
        df = _get_seo_test_data('client_secrets.json',
                                    'https://example.com',
                                    '2021-07-17',
                                    14)

        # Page level test
        filters = [{
            'filters':[{
                'dimension':'page',
                'expression': 'https://example.com/hello'
            }]
        }]

        df = _get_seo_test_data('client_secrets.json',
                                    'https://example.com',
                                    '2021-07-17',
                                    14,
                                    filters)

        # Query level test
        filters = [{
            'filters':[{
                'dimension':'query',
                'expression': 'marketing'
            }]
        }]

        df = _get_seo_test_data('client_secrets.json',
                                    'https://example.com',
                                    '2021-07-17',
                                    14,
                                    filters)
    """

    # Get the dates of the pre- and post-periods
    pre_period, post_period = _get_pre_and_post_periods(post_period_start_date, days)

    # Create basic payload
    payload = {
        'startDate': pre_period[0],
        'endDate': post_period[1],
        'dimensions': ['date'],
        'rowLimit': 10000,
        'startRow': 0,
    }

    # Add filters to the payload if provided
    if filters:
        payload['dimensionFilterGroups'] = filters

    # Run Google Search Console query using payload
    df = seo.query_google_search_console(key, site_url, payload)
    df.sort_values(by='date', ascending=True).head()
    df = df.set_index('date')

    return df


def seo_test(key,
             site_url,
             post_period_start_date,
             days,
             filters=None,
             metric='clicks'):
    """Run a simple marketing or SEO test using CausalImpact.

    Args:
        key (string): Filepath of Google Search Console API client secrets JSON keyfile
        site_url (string): Google Search Console property URL to query
        post_period_start_date (string): Start date for test in YYYY-MM-DD format
        days (int): Number of days to include in test period
        filters (optional, list): Optional list of GSC formatted query filters.
        metric (optional): Select a specific metric to examine. Default is clicks.

    Returns:
        model (object): Returns a Causal Impact model object.

    """

    # Get the SEO test data
    df = _get_seo_test_data(key, site_url, post_period_start_date, days, filters)

    # Get the dates of the pre- and post-periods
    pre_period, post_period = _get_pre_and_post_periods(post_period_start_date, days)

    # Fit the test model
    model = CausalImpact(df[metric], pre_period, post_period)

    return model




import pandas as pd
import numpy as np
from ecommercetools import transactions


def transactions_report(df, frequency='M'):
    """Create an transactions report based on a specified reporting frequency.

    Args:
        df (dataframe): Pandas dataframe of transaction items.
        frequency (optional, string, default 'M'): Optional frequency indicator (Y, Q, M, W, D)

    Returns:
        df (dataframe): Pandas dataframe of aggregated data for the specified frequency.
    """

    df['year'] = df['order_date'].dt.year
    df['quarter'] = df['order_date'].dt.quarter
    df['year_quarter'] = df['year'].astype(str) + '-' + df['quarter'].astype(str)
    df['month'] = df['order_date'].dt.month
    df['year_month'] = df['order_date'].dt.strftime('%Y-%m')
    df['week'] = df['order_date'].dt.strftime('%W')
    df['year_week'] = df['order_date'].dt.strftime('%Y-%W')
    df['day'] = df['order_date'].dt.strftime('%j')
    df['year_day'] = df['order_date'].dt.strftime('%Y-%j')

    if frequency == 'Y':
        group = 'year'
    elif frequency == 'Q':
        group = 'year_quarter'
    elif frequency == 'W':
        group = 'year_week'
    elif frequency == 'D':
        group = 'year_day'
    else:
        group = 'year_month'

    df_agg = df.groupby(group).agg(
        customers=('customer_id', 'nunique'),
        orders=('order_id', 'nunique'),
        revenue=('line_price', 'sum'),
        skus=('sku', 'count'),
        units=('quantity', 'sum')
    ).reset_index()

    df_agg['avg_order_value'] = round(df_agg['revenue'] / df_agg['orders'], 2)
    df_agg['avg_skus_per_order'] = round(df_agg['skus'] / df_agg['orders'], 2)
    df_agg['avg_units_per_order'] = round(df_agg['units'] / df_agg['orders'], 2)
    df_agg['avg_revenue_per_customer'] = round(df_agg['revenue'] / df_agg['customers'], 2)

    return df_agg


def customers_report(transaction_items_df, frequency='M'):
    """Create a customers report based on a specified reporting frequency.

    Args:
        df (dataframe): Pandas dataframe of transaction items.
        frequency (optional, string, default 'M'): Optional frequency indicator (Y, Q, M, W, D)

    Returns:
        df (dataframe): Pandas dataframe of aggregated data for the specified frequency.
    """

    df = transactions.get_transactions(transaction_items_df)

    df['period'] = df['order_date'].dt.strftime('%B, Y')
    df['year'] = df['order_date'].dt.year
    df['quarter'] = df['order_date'].dt.quarter
    df['year_quarter'] = df['year'].astype(str) + '-' + df['quarter'].astype(str)
    df['month'] = df['order_date'].dt.month
    df['year_month'] = df['order_date'].dt.strftime('%Y-%m')
    df['week'] = df['order_date'].dt.strftime('%W')
    df['year_week'] = df['order_date'].dt.strftime('%Y-%W')
    df['day'] = df['order_date'].dt.strftime('%j')
    df['year_day'] = df['order_date'].dt.strftime('%Y-%j')

    if frequency == 'Y':
        group = 'year'
    elif frequency == 'Q':
        group = 'year_quarter'
    elif frequency == 'W':
        group = 'year_week'
    elif frequency == 'D':
        group = 'year_day'
    else:
        group = 'year_month'

    df['new_customers'] = np.where(df['order_number'] == 1, 1, 0)

    df_agg = df.groupby(group).agg(
        orders=('order_id', 'nunique'),
        customers=('customer_id', 'nunique'),
        new_customers=('new_customers', 'sum'),
    ).reset_index()

    df_agg['returning_customers'] = df_agg['customers'] - df_agg['new_customers']
    df_agg['acquisition_rate'] = round((df_agg['new_customers'] / df_agg['customers']) * 100, 2)

    return df_agg
